
define_custom_obj_fields({
    oDryBonesTurnDirection = 's32',
    oDryBonesTargetYaw = 'u32',
    oDryBonesWallTimer = 'u32',
})

E_MODEL_DRY_BONES = smlua_model_util_get_id("dry_bones_geo")
local bhvDryBonesActions = { }
local bhvDryBonesAnims = { }

function start_custom_anim(obj, anim_name)
    smlua_anim_util_set_animation(obj, anim_name)
    local animInfo = obj.header.gfx.animInfo
    animInfo.animFrame = 1
    animInfo.animAccel = 0
    animInfo.animYTrans = 0
end

function bhv_dry_bones_set_action(obj, action)
    obj.oAction = action
    start_custom_anim(obj, bhvDryBonesAnims[action])
end

function bhv_dry_bones_init(obj)
    obj.oFlags = (OBJ_FLAG_COMPUTE_ANGLE_TO_MARIO | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
    obj.oMoveFlags = obj.oMoveFlags | OBJ_MOVE_ON_GROUND
    object_drop_to_floor(obj)
    cur_obj_set_home_once()
    object_set_physics(obj, 50, -400, 0, 0, 1000, 200)
    obj_scale(obj, 1.5)

    obj_set_model_extended(obj, E_MODEL_DRY_BONES)
    obj.oAction = 1
    cur_obj_hide()

    -- hitbox
    obj.oInteractType = INTERACT_BOUNCE_TOP
    obj.oHealth = 0
    obj.oNumLootCoins = 1
    obj.oIntangibleTimer = 0
    obj.hitboxRadius = obj.header.gfx.scale.x * 50
    obj.hitboxHeight = obj.header.gfx.scale.y * 50
    obj.hurtboxRadius = obj.header.gfx.scale.x * 42
    obj.hurtboxHeight = obj.header.gfx.scale.y * 40
    obj.hitboxDownOffset = obj.header.gfx.scale.y * 0
    obj.oDamageOrCoinValue = 1

    bhvDryBonesActions = {
        bhv_dry_bones_waiting,
        bhv_dry_bones_rise,
        bhv_dry_bones_idle,
        bhv_dry_bones_wander,
        bhv_dry_bones_chase,
        bhv_dry_bones_hurt,
    }

    bhvDryBonesAnims = {
        nil,
        ANIM_DRY_BONES_SPAWN,
        ANIM_DRY_BONES_IDLE,
        ANIM_DRY_BONES_WALKING,
        ANIM_DRY_BONES_WALKING,
        ANIM_DRY_BONES_HURT,
    }

    network_init_object(obj, true, {
        'oDryBonesTurnDirection',
        'oDryBonesTargetYaw',
        'oDryBonesWallTimer',
    })
end

function bhv_dry_bones_waiting(obj)
    obj.oIntangibleTimer = 2
    cur_obj_hide()
    local m = nearest_mario_state_to_object(obj)
    if m ~= nil and m.marioObj ~= nil then
        local dist = dist_between_objects(obj, m.marioObj)
        if dist < 30 * 20 then
            bhv_dry_bones_set_action(obj, 2)
        end
    end
end

function bhv_dry_bones_rise(obj)
    if obj.oTimer < 15 then
        obj.oIntangibleTimer = 2
    end
    if obj.oTimer == 1 then
        cur_obj_play_sound_2(SOUND_GENERAL_MOVING_IN_SAND)
    end
    local animFrame = obj.header.gfx.animInfo.animFrame
    local lastFrame = obj.header.gfx.animInfo.curAnim.loopEnd
    if animFrame >= (lastFrame - 1) then
        bhv_dry_bones_set_action(obj, 4)
    end
end

function bhv_dry_bones_idle(obj)
    obj_forward_vel_approach(0, 0.8)
    if obj.oTimer > 30 * 4 then
        bhv_dry_bones_set_action(obj, 4)
    end
    bhv_dry_bones_should_chase(obj)

    if bhv_dry_bones_should_chase(obj) then
        bhv_dry_bones_set_action(obj, 5)
    end
end

function bhv_dry_bones_play_footstep(obj, modulo)
    if (obj.oTimer % modulo) == 0 then
        cur_obj_play_sound_2(SOUND_OBJ_KOOPA_WALK)
    end    
end

function bhv_dry_bones_wander(obj)
    if obj.oTimer == 0 then
        if math.random(2) == 1 then
            obj.oDryBonesTurnDirection = 1
        else
            obj.oDryBonesTurnDirection = -1
        end
    end

    if obj.oDryBonesWallTimer <= 0 then
        obj.oDryBonesTargetYaw = obj.oFaceAngleYaw + 0x400 * obj.oDryBonesTurnDirection
        local distToHome = cur_obj_lateral_dist_from_obj_to_home(obj)
        if distToHome > 500 then
            obj.oDryBonesTargetYaw = cur_obj_angle_to_home()
        end
    end
    cur_obj_rotate_yaw_toward(obj.oDryBonesTargetYaw, 0x100)

    obj_forward_vel_approach(8, 0.5)
    if obj.oTimer > 30 * 7 then
        bhv_dry_bones_set_action(obj, 3)
    end

    if bhv_dry_bones_should_chase(obj) then
        bhv_dry_bones_set_action(obj, 5)
    end

    bhv_dry_bones_play_footstep(obj, 10)
end

function bhv_dry_bones_chase(obj)
    if not bhv_dry_bones_should_chase(obj) then
        bhv_dry_bones_set_action(obj, 4)
        return
    end

    local animInfo = obj.header.gfx.animInfo
    animInfo.animFrame = animInfo.animFrame + 1

    local player = nearest_player_to_object(obj)
    local angleToPlayer = obj_angle_to_object(obj, player)

    if obj.oDryBonesWallTimer <= 0 then
        obj.oDryBonesTargetYaw = angleToPlayer
    end
    cur_obj_rotate_yaw_toward(obj.oDryBonesTargetYaw, 0x250)

    obj_forward_vel_approach(20, 0.75)
    bhv_dry_bones_play_footstep(obj, 7)
end

function bhv_dry_bones_hurt(obj)
    obj.oDamageOrCoinValue = 0
    obj_forward_vel_approach(0, 0.8)
    local animFrame = obj.header.gfx.animInfo.animFrame
    local lastFrame = obj.header.gfx.animInfo.curAnim.loopEnd
    if obj.oTimer < 30 * 10 then
        if animFrame >= 10 then
            obj.header.gfx.animInfo.animFrame = 10
        end
    elseif animFrame < 20 then
        obj.header.gfx.animInfo.animFrame = 20
        cur_obj_play_sound_2(SOUND_GENERAL_PENDULUM_SWING)
    elseif animFrame >= (lastFrame - 1) then
        bhv_dry_bones_set_action(obj, 4)
    end
end

function bhv_dry_bones_should_chase(obj)
    local player = nearest_player_to_object(obj)
    if player == nil then return false end

    local distanceToPlayer = dist_between_objects(obj, player)
    if distanceToPlayer > 1500 then return false end

    local angleToPlayer = obj_angle_to_object(obj, player)
    if abs_angle_diff(obj.oMoveAngleYaw, angleToPlayer) >= 10000 then return false end

    local distToHome = cur_obj_lateral_dist_from_obj_to_home(obj)
    if distToHome > 2000 then return false end

    return true
end

function bhv_dry_bones_loop(obj)
    local animName = bhvDryBonesAnims[obj.oAction]
    if smlua_anim_util_get_current_animation_name(obj) ~= animName then
        start_custom_anim(obj, animName)
    end

    cur_obj_unhide()
    obj.oDamageOrCoinValue = 1
    local func = bhvDryBonesActions[obj.oAction]
    func(obj)

    if obj.oDryBonesWallTimer > 0 then
        obj.oDryBonesWallTimer = obj.oDryBonesWallTimer - 1
    end

    cur_obj_update_floor_and_walls()
    if (obj.oMoveFlags & OBJ_MOVE_HIT_WALL) ~= 0 then
        obj.oDryBonesWallTimer = 30 * 2
        obj.oDryBonesTargetYaw = obj.oWallAngle
    end

    cur_obj_move_standard(30)

    if (obj.oInteractStatus & INT_STATUS_INTERACTED) ~= 0 then
        obj.oInteractStatus = obj.oInteractStatus & (~INT_STATUS_INTERACTED)
        if obj.oAction ~= 6 then
            obj.oTimer = 0
            bhv_dry_bones_set_action(obj, 6)
            cur_obj_play_sound_2(SOUND_OBJ_KOOPA_DAMAGE)
        end
    end
end

id_bhvDryBones = hook_behavior(nil, OBJ_LIST_PUSHABLE, true, bhv_dry_bones_init, bhv_dry_bones_loop, 'id_bhvDryBones')
